@@ -1,6 +1,9 @@
 /*
-   This file is the C part of the bitarray package.  Almost all
-   functionality is implemented here.
+   Copyright (c) 2008 - 2019, Ilan Schnell
+   bitarray is published under the PSF license.
+
+   This file is the C part of the bitarray package.
+   All functionality is implemented here.
 
    Author: Ilan Schnell
 */
@@ -23,32 +26,13 @@
 #define Py_TPFLAGS_HAVE_WEAKREFS  0
 #endif
 
-#if PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION < 5
-/* Py_ssize_t was introduced in Python 2.5, substitute long for it */
-typedef long Py_ssize_t;
-#define PY_SSIZE_T_MAX  LONG_MAX
-#define PY_SSIZE_T_MIN  LONG_MIN
-Py_ssize_t PyNumber_AsSsize_t(PyObject *o, PyObject *exc)
-{
-    return PyLong_AsLong(o);
-}
-int PyIndex_Check(PyObject *o)
-{
-    return 0;
-}
-#define PY_SSIZE_T_FMT  "l"
-#else
-/* Python 2.5 and up uses 'n' as the format char for Py_ssize_t */
-#define PY_SSIZE_T_FMT  "n"
-#endif
-
 #if PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION < 6
 /* backward compatibility with Python 2.5 */
 #define Py_TYPE(ob)   (((PyObject *) (ob))->ob_type)
 #define Py_SIZE(ob)   (((PyVarObject *) (ob))->ob_size)
 #endif
 
-#if PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION == 7
+#if PY_MAJOR_VERSION == 3 || (PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION == 7)
 /* (new) buffer protocol */
 #define WITH_BUFFER
 #endif
@@ -1056,7 +1040,7 @@ bitarray_search(bitarrayobject *self, PyObject *args)
     bitarrayobject *xa;
     idx_t p;
 
-    if (!PyArg_ParseTuple(args, "O|" PY_SSIZE_T_FMT ":_search", &x, &limit))
+    if (!PyArg_ParseTuple(args, "O|n:_search", &x, &limit))
         return NULL;
 
     if (!bitarray_Check(x)) {
@@ -1444,7 +1428,7 @@ bitarray_fromfile(bitarrayobject *self, PyObject *args)
     idx_t t, p;
     long cur;
 
-    if (!PyArg_ParseTuple(args, "O|" PY_SSIZE_T_FMT ":fromfile", &f, &nbytes))
+    if (!PyArg_ParseTuple(args, "O|n:fromfile", &f, &nbytes))
         return NULL;
 
     fp = PyFile_AsFile(f);
@@ -2116,12 +2100,43 @@ BITWISE_IFUNC(xor)
 
 /******************* variable length encoding and decoding ***************/
 
+static int
+check_codedict(PyObject *codedict)
+{
+    PyObject *key, *value;
+    Py_ssize_t pos = 0;
+
+    if (!PyDict_Check(codedict)) {
+        PyErr_SetString(PyExc_TypeError, "dict expected");
+        return -1;
+    }
+    if (PyDict_Size(codedict) == 0) {
+        PyErr_SetString(PyExc_ValueError, "prefix code dict empty");
+        return -1;
+    }
+    while (PyDict_Next(codedict, &pos, &key, &value)) {
+        if (!bitarray_Check(value)) {
+            PyErr_SetString(PyExc_TypeError,
+                            "bitarray expected for dict value");
+            return -1;
+        }
+        if (((bitarrayobject *) value)->nbits == 0) {
+            PyErr_SetString(PyExc_ValueError, "non-empty bitarray expected");
+            return -1;
+        }
+    }
+    return 0;
+}
+
 static PyObject *
 bitarray_encode(bitarrayobject *self, PyObject *args)
 {
     PyObject *codedict, *iterable, *iter, *symbol, *bits;
 
-    if (!PyArg_ParseTuple(args, "OO:_encode", &codedict, &iterable))
+    if (!PyArg_ParseTuple(args, "OO:encode", &codedict, &iterable))
+        return NULL;
+
+    if (check_codedict(codedict) < 0)
         return NULL;
 
     iter = PyObject_GetIter(iterable);
@@ -2134,7 +2149,8 @@ bitarray_encode(bitarrayobject *self, PyObject *args)
         bits = PyDict_GetItem(codedict, symbol);
         Py_DECREF(symbol);
         if (bits == NULL) {
-            PyErr_SetString(PyExc_ValueError, "symbol not in prefix code");
+            PyErr_SetString(PyExc_ValueError,
+                            "symbol not defined in prefix code");
             goto error;
         }
         if (extend_bitarray(self, (bitarrayobject *) bits) < 0)
@@ -2150,72 +2166,194 @@ error:
 }
 
 PyDoc_STRVAR(encode_doc,
-"_encode(code, iterable)\n\
+"encode(code, iterable)\n\
 \n\
-like the encode method without code checking");
+Given a prefix code (a dict mapping symbols to bitarrays),\n\
+iterate over the iterable object with symbols, and extend the bitarray\n\
+with the corresponding bitarray for each symbols.");
 
 
-/* return the leave node resulting from traversing the (binary) tree,
-   or, when the iteration is finished, NULL
-*/
-static PyObject *
-tree_traverse(bitarrayobject *self, idx_t *indexp, PyObject *tree)
+/* Binary tree definition */
+typedef struct _bin_node
 {
-    PyObject *subtree;
-    long vi;
+    struct _bin_node *child[2];
+    PyObject *symbol;
+} binode;
 
-    if (*indexp == self->nbits)  /* stop iterator */
+
+static binode *
+new_binode(void)
+{
+    binode *nd;
+
+    nd = PyMem_Malloc(sizeof *nd);
+    if (nd == NULL) {
+        PyErr_NoMemory();
         return NULL;
-
-    vi = GETBIT(self, *indexp);
-    (*indexp)++;
-    subtree = PyList_GetItem(tree, vi);
-
-    if (PyList_Check(subtree) && PyList_Size(subtree) == 2)
-        return tree_traverse(self, indexp, subtree);
-    else
-        return subtree;
+    }
+    nd->child[0] = NULL;
+    nd->child[1] = NULL;
+    nd->symbol = NULL;
+    return nd;
 }
 
-#define IS_EMPTY_LIST(x)  (PyList_Check(x) && PyList_Size(x) == 0)
+static void
+delete_binode_tree(binode *tree)
+{
+    if (tree == NULL)
+        return;
+
+    delete_binode_tree(tree->child[0]);
+    delete_binode_tree(tree->child[1]);
+    PyMem_Free(tree);
+}
+
+static int
+insert_symbol(binode *tree, bitarrayobject *ba, PyObject *symbol)
+{
+    binode *nd = tree, *prev;
+    Py_ssize_t i;
+    int k;
+
+    for (i = 0; i < ba->nbits; i++) {
+        k = GETBIT(ba, i);
+        prev = nd;
+        nd = nd->child[k];
+
+        /* we cannot have already a symbol when branching to the new leaf */
+        if (nd && nd->symbol)
+            goto ambiguity;
+
+        if (!nd) {
+            nd = new_binode();
+            if (nd == NULL)
+                return -1;
+            prev->child[k] = nd;
+        }
+    }
+    /* the new leaf node cannot already have a symbol or children */
+    if (nd->symbol || nd->child[0] || nd->child[1])
+        goto ambiguity;
+
+    nd->symbol = symbol;
+    return 0;
+
+ ambiguity:
+    PyErr_SetString(PyExc_ValueError, "prefix code ambiguous");
+    return -1;
+}
+
+static binode *
+make_tree(PyObject *codedict)
+{
+    binode *tree;
+    PyObject *symbol, *array;
+    Py_ssize_t pos = 0;
+
+    tree = new_binode();
+    if (tree == NULL)
+        return NULL;
+
+    while (PyDict_Next(codedict, &pos, &symbol, &array)) {
+        if (insert_symbol(tree, (bitarrayobject *) array, symbol) < 0) {
+            delete_binode_tree(tree);
+            return NULL;
+        }
+    }
+    return tree;
+}
+
+/*
+  Traverse tree using the branches corresponding to the bitarray `ba`,
+  starting at *indexp, and return the symbol at the leaf node (or NULL
+  when the end of the bitarray has been reached).
+*/
+static PyObject *
+traverse_tree(binode *tree, bitarrayobject *ba, idx_t *indexp)
+{
+    binode *nd = tree;
+    int k;
+
+    while (*indexp < ba->nbits) {
+        k = GETBIT(ba, *indexp);
+        (*indexp)++;
+        nd = nd->child[k];
+        if (nd == NULL) {
+            PyErr_SetString(PyExc_ValueError,
+                            "prefix code does not match data in bitarray");
+            return NULL;
+        }
+        if (nd->symbol)  /* leaf */
+            return nd->symbol;
+    }
+    if (nd != tree)
+        PyErr_SetString(PyExc_ValueError, "decoding not terminated");
+
+    return NULL;
+}
 
 static PyObject *
-bitarray_decode(bitarrayobject *self, PyObject *tree)
+bitarray_decode(bitarrayobject *self, PyObject *codedict)
 {
-    PyObject *symbol, *list;
-    idx_t index = 0;
+    binode *tree, *nd;
+    PyObject *list;
+    Py_ssize_t i;
+    int k;
 
-    list = PyList_New(0);
-    if (list == NULL)
+    if (check_codedict(codedict) < 0)
         return NULL;
-    /* traverse binary tree and append symbols to the result list */
-    while ((symbol = tree_traverse(self, &index, tree)) != NULL) {
-        if (IS_EMPTY_LIST(symbol)) {
+
+    tree = make_tree(codedict);
+    if (tree == NULL || PyErr_Occurred())
+        return NULL;
+
+    nd = tree;
+    list = PyList_New(0);
+    if (list == NULL) {
+        delete_binode_tree(tree);
+        return NULL;
+    }
+    /* traverse tree (just like above) */
+    for (i = 0; i < self->nbits; i++) {
+        k = GETBIT(self, i);
+        nd = nd->child[k];
+        if (nd == NULL) {
             PyErr_SetString(PyExc_ValueError,
                             "prefix code does not match data in bitarray");
             goto error;
         }
-        if (PyList_Append(list, symbol) < 0)
-            goto error;
+        if (nd->symbol) {  /* leaf */
+            if (PyList_Append(list, nd->symbol) < 0)
+                goto error;
+            nd = tree;
+        }
     }
+    if (nd != tree) {
+        PyErr_SetString(PyExc_ValueError, "decoding not terminated");
+        goto error;
+    }
+    delete_binode_tree(tree);
     return list;
+
 error:
+    delete_binode_tree(tree);
     Py_DECREF(list);
     return NULL;
 }
 
 PyDoc_STRVAR(decode_doc,
-"_decode(tree) -> list\n\
+"decode(code) -> list\n\
 \n\
-Given a tree, decode the content of the bitarray and return the list of\n\
-symbols.");
+Given a prefix code (a dict mapping symbols to bitarrays),\n\
+decode the content of the bitarray and return it as a list of symbols.");
 
 /*********************** (Bitarray) Decode Iterator *********************/
 
+
 typedef struct {
     PyObject_HEAD
     bitarrayobject *bao;        /* bitarray we're searching in */
-    PyObject *tree;             /* prefix tree containing symbols */
+    binode *tree;               /* prefix tree containing symbols */
     idx_t index;                /* current index in bitarray */
 } decodeiterobject;
 
@@ -2223,30 +2361,41 @@ static PyTypeObject DecodeIter_Type;
 
 #define DecodeIter_Check(op)  PyObject_TypeCheck(op, &DecodeIter_Type)
 
+
+
 /* create a new initialized bitarray search iterator object */
 static PyObject *
-bitarray_iterdecode(bitarrayobject *self, PyObject *tree)
+bitarray_iterdecode(bitarrayobject *self, PyObject *codedict)
 {
     decodeiterobject *it;  /* iterator to be returned */
+    binode *tree;
+
+    if (check_codedict(codedict) < 0)
+        return NULL;
+
+    tree = make_tree(codedict);
+    if (tree == NULL || PyErr_Occurred())
+        return NULL;
 
     it = PyObject_GC_New(decodeiterobject, &DecodeIter_Type);
     if (it == NULL)
         return NULL;
 
+    it->tree = tree;
+
     Py_INCREF(self);
     it->bao = self;
-    Py_INCREF(tree);
-    it->tree = tree;
     it->index = 0;
     PyObject_GC_Track(it);
     return (PyObject *) it;
 }
 
 PyDoc_STRVAR(iterdecode_doc,
-"_iterdecode(tree) -> iterator\n\
+"iterdecode(code) -> iterator\n\
 \n\
-Given a tree, decode the content of the bitarray and iterate over the\n\
-symbols.");
+Given a prefix code (a dict mapping symbols to bitarrays),\n\
+decode the content of the bitarray and return an iterator over\n\
+the symbols.");
 
 static PyObject *
 decodeiter_next(decodeiterobject *it)
@@ -2254,14 +2403,9 @@ decodeiter_next(decodeiterobject *it)
     PyObject *symbol;
 
     assert(DecodeIter_Check(it));
-    symbol = tree_traverse(it->bao, &(it->index), it->tree);
-    if (symbol == NULL)  /* stop iteration */
+    symbol = traverse_tree(it->tree, it->bao, &(it->index));
+    if (symbol == NULL)  /* stop iteration OR error occured */
         return NULL;
-    if (IS_EMPTY_LIST(symbol)) {
-        PyErr_SetString(PyExc_ValueError,
-                        "prefix code does not match data in bitarray");
-        return NULL;
-    }
     Py_INCREF(symbol);
     return symbol;
 }
@@ -2269,9 +2413,9 @@ decodeiter_next(decodeiterobject *it)
 static void
 decodeiter_dealloc(decodeiterobject *it)
 {
+    delete_binode_tree(it->tree);
     PyObject_GC_UnTrack(it);
     Py_XDECREF(it->bao);
-    Py_XDECREF(it->tree);
     PyObject_GC_Del(it);
 }
 
@@ -2284,7 +2428,7 @@ decodeiter_traverse(decodeiterobject *it, visitproc visit, void *arg)
 
 static PyTypeObject DecodeIter_Type = {
 #ifdef IS_PY3K
-    PyVarObject_HEAD_INIT(&DecodeIter_Type, 0)
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
 #else
     PyObject_HEAD_INIT(NULL)
     0,                                        /* ob_size */
@@ -2399,7 +2543,7 @@ searchiter_traverse(searchiterobject *it, visitproc visit, void *arg)
 
 static PyTypeObject SearchIter_Type = {
 #ifdef IS_PY3K
-    PyVarObject_HEAD_INIT(&SearchIter_Type, 0)
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
 #else
     PyObject_HEAD_INIT(NULL)
     0,                                        /* ob_size */
@@ -2452,11 +2596,11 @@ bitarray_methods[] = {
      copy_doc},
     {"count",        (PyCFunction) bitarray_count,       METH_VARARGS,
      count_doc},
-    {"_decode",      (PyCFunction) bitarray_decode,      METH_O,
+    {"decode",       (PyCFunction) bitarray_decode,      METH_O,
      decode_doc},
-    {"_iterdecode",  (PyCFunction) bitarray_iterdecode,  METH_O,
+    {"iterdecode",   (PyCFunction) bitarray_iterdecode,  METH_O,
      iterdecode_doc},
-    {"_encode",      (PyCFunction) bitarray_encode,      METH_VARARGS,
+    {"encode",       (PyCFunction) bitarray_encode,      METH_VARARGS,
      encode_doc},
     {"endian",       (PyCFunction) bitarray_endian,      METH_NOARGS,
      endian_doc},
@@ -2763,7 +2907,7 @@ bitarrayiter_traverse(bitarrayiterobject *it, visitproc visit, void *arg)
 
 static PyTypeObject BitarrayIter_Type = {
 #ifdef IS_PY3K
-    PyVarObject_HEAD_INIT(&BitarrayIter_Type, 0)
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
 #else
     PyObject_HEAD_INIT(NULL)
     0,                                        /* ob_size */
@@ -2800,6 +2944,8 @@ static PyTypeObject BitarrayIter_Type = {
 
 /********************* Bitarray Buffer Interface ************************/
 #ifdef WITH_BUFFER
+
+#if PY_MAJOR_VERSION == 2
 static Py_ssize_t
 bitarray_buffer_getreadbuf(bitarrayobject *self,
                            Py_ssize_t index, const void **ptr)
@@ -2844,6 +2990,8 @@ bitarray_buffer_getcharbuf(bitarrayobject *self,
     return Py_SIZE(self);
 }
 
+#endif
+
 static int
 bitarray_getbuffer(bitarrayobject *self, Py_buffer *view, int flags)
 {
@@ -2870,19 +3018,23 @@ bitarray_releasebuffer(bitarrayobject *self, Py_buffer *view)
 }
 
 static PyBufferProcs bitarray_as_buffer = {
+#if PY_MAJOR_VERSION == 2   /* old buffer protocol */
     (readbufferproc) bitarray_buffer_getreadbuf,
     (writebufferproc) bitarray_buffer_getwritebuf,
     (segcountproc) bitarray_buffer_getsegcount,
     (charbufferproc) bitarray_buffer_getcharbuf,
+#endif
     (getbufferproc) bitarray_getbuffer,
     (releasebufferproc) bitarray_releasebuffer,
 };
+
 #endif  /* WITH_BUFFER */
+
 /************************** Bitarray Type *******************************/
 
 static PyTypeObject Bitarraytype = {
 #ifdef IS_PY3K
-    PyVarObject_HEAD_INIT(&Bitarraytype, 0)
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
 #else
     PyObject_HEAD_INIT(NULL)
     0,                                        /* ob_size */
@@ -2911,7 +3063,7 @@ static PyTypeObject Bitarraytype = {
     0,                                        /* tp_as_buffer */
 #endif
     Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_WEAKREFS
-#ifdef WITH_BUFFER
+#if defined(WITH_BUFFER) && PY_MAJOR_VERSION == 2
     | Py_TPFLAGS_HAVE_NEWBUFFER
 #endif
     ,                                         /* tp_flags */
@@ -3047,6 +3199,8 @@ init_bitarray(void)
     PyObject *m;
 
     Py_TYPE(&Bitarraytype) = &PyType_Type;
+    Py_TYPE(&SearchIter_Type) = &PyType_Type;
+    Py_TYPE(&DecodeIter_Type) = &PyType_Type;
     Py_TYPE(&BitarrayIter_Type) = &PyType_Type;
 #ifdef IS_PY3K
     m = PyModule_Create(&moduledef);
